from .large_graph_indexer import LargeGraphIndexer
from .rag_loader import RAGQueryLoader
from .utils import *  # noqa
from .models import *  # noqa

__all__ = classes = [
    'LargeGraphIndexer',
    'RAGQueryLoader',
]



from functools import wraps
from typing import Optional

import torch

CUDA_PROFILE_STARTED = False


def begin_cuda_profile():
    global CUDA_PROFILE_STARTED
    prev_state = CUDA_PROFILE_STARTED
    if prev_state is False:
        CUDA_PROFILE_STARTED = True
        torch.cuda.cudart().cudaProfilerStart()
    return prev_state


def end_cuda_profile(prev_state: bool):
    global CUDA_PROFILE_STARTED
    CUDA_PROFILE_STARTED = prev_state
    if prev_state is False:
        torch.cuda.cudart().cudaProfilerStop()


def nvtxit(name: Optional[str] = None, n_warmups: int = 0,
           n_iters: Optional[int] = None):
    """Enables NVTX profiling for a function.

    Args:
        name (Optional[str], optional): Name to give the reference frame for
            the function being wrapped. Defaults to the name of the
            function in code.
        n_warmups (int, optional): Number of iters to call that function
            before starting. Defaults to 0.
        n_iters (Optional[int], optional): Number of iters of that function to
            record. Defaults to all of them.
    """
    def nvtx(func):

        nonlocal name
        iters_so_far = 0
        if name is None:
            name = func.__name__

        @wraps(func)
        def wrapper(*args, **kwargs):
            nonlocal iters_so_far
            if not torch.cuda.is_available():
                return func(*args, **kwargs)
            elif iters_so_far < n_warmups:
                iters_so_far += 1
                return func(*args, **kwargs)
            elif n_iters is None or iters_so_far < n_iters + n_warmups:
                prev_state = begin_cuda_profile()
                torch.cuda.nvtx.range_push(f"{name}_{iters_so_far}")
                result = func(*args, **kwargs)
                torch.cuda.nvtx.range_pop()
                end_cuda_profile(prev_state)
                iters_so_far += 1
                return result
            else:
                return func(*args, **kwargs)

        return wrapper

    return nvtx



import math
from typing import Optional

import torch
import torch.nn.functional as F
from torch import Tensor
from torch.nn import Parameter

from torch_geometric.nn.conv import MessagePassing
from torch_geometric.nn.conv.gcn_conv import gcn_norm
from torch_geometric.nn.inits import kaiming_uniform, uniform
from torch_geometric.typing import Adj, OptPairTensor, OptTensor, SparseTensor


class Linear(torch.nn.Module):
    def __init__(self, in_channels, out_channels, groups=1, bias=True):
        super().__init__()
        assert in_channels % groups == 0 and out_channels % groups == 0

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.groups = groups

        self.weight = Parameter(
            torch.empty(groups, in_channels // groups, out_channels // groups))

        if bias:
            self.bias = Parameter(torch.empty(out_channels))
        else:
            self.register_parameter('bias', None)

        self.reset_parameters()

    def reset_parameters(self):
        kaiming_uniform(self.weight, fan=self.weight.size(1), a=math.sqrt(5))
        uniform(self.weight.size(1), self.bias)

    def forward(self, src):
        # Input: [*, in_channels]
        # Output: [*, out_channels]

        if self.groups > 1:
            size = src.size()[:-1]
            src = src.view(-1, self.groups, self.in_channels // self.groups)
            src = src.transpose(0, 1).contiguous()
            out = torch.matmul(src, self.weight)
            out = out.transpose(1, 0).contiguous()
            out = out.view(size + (self.out_channels, ))
        else:
            out = torch.matmul(src, self.weight.squeeze(0))

        if self.bias is not None:
            out = out + self.bias

        return out

    def __repr__(self) -> str:  # pragma: no cover
        return (f'{self.__class__.__name__}({self.in_channels}, '
                f'{self.out_channels}, groups={self.groups})')


def restricted_softmax(src, dim: int = -1, margin: float = 0.):
    src_max = torch.clamp(src.max(dim=dim, keepdim=True)[0], min=0.)
    out = (src - src_max).exp()
    out = out / (out.sum(dim=dim, keepdim=True) + (margin - src_max).exp())
    return out


class Attention(torch.nn.Module):
    def __init__(self, dropout=0):
        super().__init__()
        self.dropout = dropout

    def forward(self, query, key, value):
        return self.compute_attention(query, key, value)

    def compute_attention(self, query, key, value):
        # query: [*, query_entries, dim_k]
        # key: [*, key_entries, dim_k]
        # value: [*, key_entries, dim_v]
        # Output: [*, query_entries, dim_v]

        assert query.dim() == key.dim() == value.dim() >= 2
        assert query.size(-1) == key.size(-1)
        assert key.size(-2) == value.size(-2)

        # Score: [*, query_entries, key_entries]
        score = torch.matmul(query, key.transpose(-2, -1))
        score = score / math.sqrt(key.size(-1))
        score = restricted_softmax(score, dim=-1)
        score = F.dropout(score, p=self.dropout, training=self.training)

        return torch.matmul(score, value)

    def __repr__(self) -> str:  # pragma: no cover
        return f'{self.__class__.__name__}(dropout={self.dropout})'


class MultiHead(Attention):
    def __init__(self, in_channels, out_channels, heads=1, groups=1, dropout=0,
                 bias=True):
        super().__init__(dropout)

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.heads = heads
        self.groups = groups
        self.bias = bias

        assert in_channels % heads == 0 and out_channels % heads == 0
        assert in_channels % groups == 0 and out_channels % groups == 0
        assert max(groups, self.heads) % min(groups, self.heads) == 0

        self.lin_q = Linear(in_channels, out_channels, groups, bias)
        self.lin_k = Linear(in_channels, out_channels, groups, bias)
        self.lin_v = Linear(in_channels, out_channels, groups, bias)

        self.reset_parameters()

    def reset_parameters(self):
        self.lin_q.reset_parameters()
        self.lin_k.reset_parameters()
        self.lin_v.reset_parameters()

    def forward(self, query, key, value):
        # query: [*, query_entries, in_channels]
        # key: [*, key_entries, in_channels]
        # value: [*, key_entries, in_channels]
        # Output: [*, query_entries, out_channels]

        assert query.dim() == key.dim() == value.dim() >= 2
        assert query.size(-1) == key.size(-1) == value.size(-1)
        assert key.size(-2) == value.size(-2)

        query = self.lin_q(query)
        key = self.lin_k(key)
        value = self.lin_v(value)

        # query: [*, heads, query_entries, out_channels // heads]
        # key: [*, heads, key_entries, out_channels // heads]
        # value: [*, heads, key_entries, out_channels // heads]
        size = query.size()[:-2]
        out_channels_per_head = self.out_channels // self.heads

        query_size = size + (query.size(-2), self.heads, out_channels_per_head)
        query = query.view(query_size).transpose(-2, -3)

        key_size = size + (key.size(-2), self.heads, out_channels_per_head)
        key = key.view(key_size).transpose(-2, -3)

        value_size = size + (value.size(-2), self.heads, out_channels_per_head)
        value = value.view(value_size).transpose(-2, -3)

        # Output: [*, heads, query_entries, out_channels // heads]
        out = self.compute_attention(query, key, value)
        # Output: [*, query_entries, heads, out_channels // heads]
        out = out.transpose(-3, -2).contiguous()
        # Output: [*, query_entries, out_channels]
        out = out.view(size + (query.size(-2), self.out_channels))

        return out

    def __repr__(self) -> str:  # pragma: no cover
        return (f'{self.__class__.__name__}({self.in_channels}, '
                f'{self.out_channels}, heads={self.heads}, '
                f'groups={self.groups}, dropout={self.dropout}, '
                f'bias={self.bias})')


class DNAConv(MessagePassing):
    r"""The dynamic neighborhood aggregation operator from the `"Just Jump:
    Towards Dynamic Neighborhood Aggregation in Graph Neural Networks"
    <https://arxiv.org/abs/1904.04849>`_ paper.

    .. math::
        \mathbf{x}_v^{(t)} = h_{\mathbf{\Theta}}^{(t)} \left( \mathbf{x}_{v
        \leftarrow v}^{(t)}, \left\{ \mathbf{x}_{v \leftarrow w}^{(t)} : w \in
        \mathcal{N}(v) \right\} \right)

    based on (multi-head) dot-product attention

    .. math::
        \mathbf{x}_{v \leftarrow w}^{(t)} = \textrm{Attention} \left(
        \mathbf{x}^{(t-1)}_v \, \mathbf{\Theta}_Q^{(t)}, [\mathbf{x}_w^{(1)},
        \ldots, \mathbf{x}_w^{(t-1)}] \, \mathbf{\Theta}_K^{(t)}, \,
        [\mathbf{x}_w^{(1)}, \ldots, \mathbf{x}_w^{(t-1)}] \,
        \mathbf{\Theta}_V^{(t)} \right)

    with :math:`\mathbf{\Theta}_Q^{(t)}, \mathbf{\Theta}_K^{(t)},
    \mathbf{\Theta}_V^{(t)}` denoting (grouped) projection matrices for query,
    key and value information, respectively.
    :math:`h^{(t)}_{\mathbf{\Theta}}` is implemented as a non-trainable
    version of :class:`torch_geometric.nn.conv.GCNConv`.

    .. note::
        In contrast to other layers, this operator expects node features as
        shape :obj:`[num_nodes, num_layers, channels]`.

    Args:
        channels (int): Size of each input/output sample.
        heads (int, optional): Number of multi-head-attentions.
            (default: :obj:`1`)
        groups (int, optional): Number of groups to use for all linear
            projections. (default: :obj:`1`)
        dropout (float, optional): Dropout probability of attention
            coefficients. (default: :obj:`0.`)
        cached (bool, optional): If set to :obj:`True`, the layer will cache
            the computation of :math:`\mathbf{\hat{D}}^{-1/2} \mathbf{\hat{A}}
            \mathbf{\hat{D}}^{-1/2}` on first execution, and will use the
            cached version for further executions.
            This parameter should only be set to :obj:`True` in transductive
            learning scenarios. (default: :obj:`False`)
        normalize (bool, optional): Whether to add self-loops and apply
            symmetric normalization. (default: :obj:`True`)
        add_self_loops (bool, optional): If set to :obj:`False`, will not add
            self-loops to the input graph. (default: :obj:`True`)
        bias (bool, optional): If set to :obj:`False`, the layer will not learn
            an additive bias. (default: :obj:`True`)
        **kwargs (optional): Additional arguments of
            :class:`torch_geometric.nn.conv.MessagePassing`.

    Shapes:
        - **input:**
          node features :math:`(|\mathcal{V}|, L, F)` where :math:`L` is the
          number of layers,
          edge indices :math:`(2, |\mathcal{E}|)`
        - **output:** node features :math:`(|\mathcal{V}|, F)`
    """

    _cached_edge_index: Optional[OptPairTensor]
    _cached_adj_t: Optional[SparseTensor]

    def __init__(self, channels: int, heads: int = 1, groups: int = 1,
                 dropout: float = 0., cached: bool = False,
                 normalize: bool = True, add_self_loops: bool = True,
                 bias: bool = True, **kwargs):
        kwargs.setdefault('aggr', 'add')
        super().__init__(node_dim=0, **kwargs)

        self.bias = bias
        self.cached = cached
        self.normalize = normalize
        self.add_self_loops = add_self_loops

        self._cached_edge_index = None
        self._cached_adj_t = None

        self.multi_head = MultiHead(channels, channels, heads, groups, dropout,
                                    bias)
        self.reset_parameters()

    def reset_parameters(self):
        super().reset_parameters()
        self.multi_head.reset_parameters()
        self._cached_edge_index = None
        self._cached_adj_t = None

    def forward(
        self,
        x: Tensor,
        edge_index: Adj,
        edge_weight: OptTensor = None,
    ) -> Tensor:
        r"""Runs the forward pass of the module.

        Args:
            x (torch.Tensor): The input node features of shape
                :obj:`[num_nodes, num_layers, channels]`.
            edge_index (torch.Tensor or SparseTensor): The edge indices.
            edge_weight (torch.Tensor, optional): The edge weights.
                (default: :obj:`None`)
        """
        if x.dim() != 3:
            raise ValueError('Feature shape must be [num_nodes, num_layers, '
                             'channels].')

        if self.normalize:
            if isinstance(edge_index, Tensor):
                cache = self._cached_edge_index
                if cache is None:
                    edge_index, edge_weight = gcn_norm(  # yapf: disable
                        edge_index, edge_weight, x.size(self.node_dim), False,
                        self.add_self_loops, self.flow, dtype=x.dtype)
                    if self.cached:
                        self._cached_edge_index = (edge_index, edge_weight)
                else:
                    edge_index, edge_weight = cache[0], cache[1]

            elif isinstance(edge_index, SparseTensor):
                cache = self._cached_adj_t
                if cache is None:
                    edge_index = gcn_norm(  # yapf: disable
                        edge_index, edge_weight, x.size(self.node_dim), False,
                        self.add_self_loops, self.flow, dtype=x.dtype)
                    if self.cached:
                        self._cached_adj_t = edge_index
                else:
                    edge_index = cache

        # propagate_type: (x: Tensor, edge_weight: OptTensor)
        return self.propagate(edge_index, x=x, edge_weight=edge_weight)

    def message(self, x_i: Tensor, x_j: Tensor, edge_weight: Tensor) -> Tensor:
        x_i = x_i[:, -1:]  # [num_edges, 1, channels]
        out = self.multi_head(x_i, x_j, x_j)  # [num_edges, 1, channels]
        return edge_weight.view(-1, 1) * out.squeeze(1)

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}({self.multi_head.in_channels}, '
                f'heads={self.multi_head.heads}, '
                f'groups={self.multi_head.groups})')



import math
from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torch.nn import Parameter

from ...nn.conv import MessagePassing
from ...nn.dense.linear import Linear
from ...nn.inits import glorot, zeros
from ...typing import Adj, OptTensor, Tuple
from ...utils import get_ppr, is_sparse, scatter, softmax
from .basic_gnn import GCN


class LPFormer(nn.Module):
    r"""The LPFormer model from the
    `"LPFormer: An Adaptive Graph Transformer for Link Prediction"
    <https://arxiv.org/abs/2310.11009>`_ paper.

    .. note::

        For an example of using LPFormer, see
        `examples/lpformer.py
        <https://github.com/pyg-team/pytorch_geometric/blob/master/examples/
        lpformer.py>`_.

    Args:
        in_channels (int): Size of input dimension
        hidden_channels (int): Size of hidden dimension
        num_gnn_layers (int, optional): Number of GNN layers
            (default: :obj:`2`)
        gnn_dropout(float, optional): Dropout used for GNN
            (default: :obj:`0.1`)
        num_transformer_layers (int, optional): Number of Transformer layers
            (default: :obj:`1`)
        num_heads (int, optional): Number of heads to use in MHA
            (default: :obj:`1`)
        transformer_dropout (float, optional): Dropout used for Transformer
            (default: :obj:`0.1`)
        ppr_thresholds (list): PPR thresholds for different types of nodes.
            Types include (in order) common neighbors, 1-Hop nodes
            (that aren't CNs), and all other nodes.
            (default: :obj:`[0, 1e-4, 1e-2]`)
        gcn_cache (bool, optional): Whether to cache edge indices
            during message passing. (default: :obj:`False`)
    """
    def __init__(
        self,
        in_channels: int,
        hidden_channels: int,
        num_gnn_layers: int = 2,
        gnn_dropout: float = 0.1,
        num_transformer_layers: int = 1,
        num_heads: int = 1,
        transformer_dropout: float = 0.1,
        ppr_thresholds: list = None,
        gcn_cache=False,
    ):
        super().__init__()

        # Default thresholds
        if ppr_thresholds is None:
            ppr_thresholds = [0, 1e-4, 1e-2]

        if len(ppr_thresholds) == 3:
            self.thresh_cn = ppr_thresholds[0]
            self.thresh_1hop = ppr_thresholds[1]
            self.thresh_non1hop = ppr_thresholds[2]
        else:
            raise ValueError(
                "Argument 'ppr_thresholds' must only be length 3!")

        self.in_dim = in_channels
        self.hid_dim = hidden_channels
        self.gnn_drop = gnn_dropout
        self.trans_drop = transformer_dropout

        self.gnn = GCN(in_channels, hidden_channels, num_gnn_layers,
                       dropout=gnn_dropout, norm="layer_norm",
                       cached=gcn_cache)
        self.gnn_norm = nn.LayerNorm(hidden_channels)

        # Create Transformer Layers
        self.att_layers = nn.ModuleList()
        for il in range(num_transformer_layers):
            if il == 0:
                node_dim = None
                self.out_dim = self.hid_dim * 2 if num_transformer_layers > 1 \
                    else self.hid_dim
            elif il == self.num_layers - 1:
                node_dim = self.hid_dim
            else:
                self.out_dim = node_dim = self.hid_dim

            self.att_layers.append(
                LPAttLayer(self.hid_dim, self.out_dim, node_dim, num_heads,
                           self.trans_drop))

        self.elementwise_lin = MLP(self.hid_dim, self.hid_dim, self.hid_dim)

        # Relative Positional Encodings
        self.ppr_encoder_cn = MLP(2, self.hid_dim, self.hid_dim)
        self.ppr_encoder_onehop = MLP(2, self.hid_dim, self.hid_dim)
        self.ppr_encoder_non1hop = MLP(2, self.hid_dim, self.hid_dim)

        # thresh=1 implies ignoring some set of nodes
        # Also allows us to be more efficient later
        if self.thresh_non1hop == 1 and self.thresh_1hop == 1:
            self.mask = "cn"
        elif self.thresh_non1hop == 1 and self.thresh_1hop < 1:
            self.mask = "1-hop"
        else:
            self.mask = "all"

        # 4 is for counts of diff nodes
        pairwise_dim = self.hid_dim * num_heads + 4
        self.pairwise_lin = MLP(pairwise_dim, pairwise_dim, self.hid_dim)

        self.score_func = MLP(self.hid_dim * 2, self.hid_dim * 2, 1, norm=None)

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}({self.in_dim}, '
                f'{self.hid_dim}, num_gnn_layers={self.gnn.num_layers}, '
                f'num_transformer_layers={len(self.att_layers)})')

    def reset_parameters(self):
        r"""Resets all learnable parameters of the module."""
        self.gnn.reset_parameters()
        self.gnn_norm.reset_parameters()
        self.elementwise_lin.reset_parameters()
        self.pairwise_lin.reset_parameters()
        self.ppr_encoder_cn.reset_parameters()
        self.ppr_encoder_onehop.reset_parameters()
        self.ppr_encoder_non1hop.reset_parameters()
        self.score_func.reset_parameters()
        for i in range(len(self.att_layers)):
            self.att_layers[i].reset_parameters()

    def forward(
        self,
        batch: Tensor,
        x: Tensor,
        edge_index: Adj,
        ppr_matrix: Tensor,
    ) -> Tensor:
        r"""Forward Pass of LPFormer.

        Returns raw logits for each link

        Args:
            batch (Tensor): The batch vector.
                Denotes which node pairs to predict.
            x (Tensor): Input node features
            edge_index (torch.Tensor, SparseTensor): The edge indices.
                Either in COO or SparseTensor format
            ppr_matrix (Tensor): PPR matrix
        """
        batch = batch.to(x.device)

        X_node = self.propagate(x, edge_index)
        x_i, x_j = X_node[batch[0]], X_node[batch[1]]
        elementwise_edge_feats = self.elementwise_lin(x_i * x_j)

        # Ensure in sparse format
        # Need as native torch.sparse for later computations
        # (necessary operations are not supported by PyG SparseTensor)
        if not edge_index.is_sparse:
            num_nodes = ppr_matrix.size(1)
            vals = torch.ones(len(edge_index[0]), device=edge_index.device)
            edge_index = torch.sparse_coo_tensor(edge_index, vals,
                                                 [num_nodes, num_nodes])
        # Checks if SparseTensor, if so the convert
        if is_sparse(edge_index) and not edge_index.is_sparse:
            edge_index = edge_index.to_torch_sparse_coo_tensor()

        # Ensure {0, 1}
        edge_index = edge_index.coalesce().bool().int()

        pairwise_feats = self.calc_pairwise(batch, X_node, edge_index,
                                            ppr_matrix)
        combined_feats = torch.cat((elementwise_edge_feats, pairwise_feats),
                                   dim=-1)

        logits = self.score_func(combined_feats)
        return logits

    def propagate(self, x: Tensor, adj: Adj) -> Tensor:
        """Propagate via GNN.

        Args:
            x (Tensor): Node features
            adj (torch.Tensor, SparseTensor): Adjacency matrix
        """
        x = F.dropout(x, p=self.gnn_drop, training=self.training)
        X_node = self.gnn(x, adj)
        X_node = self.gnn_norm(X_node)

        return X_node

    def calc_pairwise(self, batch: Tensor, X_node: Tensor, adj_mask: Tensor,
                      ppr_matrix: Tensor) -> Tensor:
        r"""Calculate the pairwise features for the node pairs.

        Args:
            batch (Tensor): The batch vector.
                Denotes which node pairs to predict.
            X_node (Tensor): Node representations
            adj_mask (Tensor): Mask of adjacency matrix used for computing the
                different node types.
            ppr_matrix (Tensor): PPR matrix
        """
        k_i, k_j = X_node[batch[0]], X_node[batch[1]]
        pairwise_feats = torch.cat((k_i, k_j), dim=-1)

        cn_info, onehop_info, non1hop_info = self.compute_node_mask(
            batch, adj_mask, ppr_matrix)

        all_mask = cn_info[0]
        if onehop_info is not None:
            all_mask = torch.cat((all_mask, onehop_info[0]), dim=-1)
        if non1hop_info is not None:
            all_mask = torch.cat((all_mask, non1hop_info[0]), dim=-1)

        pes = self.get_pos_encodings(cn_info[1:], onehop_info[1:],
                                     non1hop_info[1:])

        for lay in range(len(self.att_layers)):
            pairwise_feats = self.att_layers[lay](all_mask, pairwise_feats,
                                                  X_node, pes)

        num_cns, num_1hop, num_non1hop, num_neigh = self.get_structure_cnts(
            batch, cn_info, onehop_info, non1hop_info)

        pairwise_feats = torch.cat(
            (pairwise_feats, num_cns, num_1hop, num_non1hop, num_neigh),
            dim=-1)

        pairwise_feats = self.pairwise_lin(pairwise_feats)
        return pairwise_feats

    def get_pos_encodings(
            self, cn_ppr: Tuple[Tensor, Tensor],
            onehop_ppr: Optional[Tuple[Tensor, Tensor]] = None,
            non1hop_ppr: Optional[Tuple[Tensor, Tensor]] = None) -> Tensor:
        r"""Calculate the PPR-based relative positional encodings.

        Due to thresholds, sometimes we don't have 1-hop or >1-hop nodes.
        In those cases, the value of onehop_ppr and/or non1hop_ppr should
        be `None`.

        Args:
            cn_ppr (tuple, optional): PPR scores of CNs.
            onehop_ppr (tuple, optional): PPR scores of 1-Hop.
                (default: :obj:`None`)
            non1hop_ppr (tuple, optional): PPR scores of >1-Hop.
                (default: :obj:`None`)
        """
        cn_a = self.ppr_encoder_cn(torch.stack((cn_ppr[0], cn_ppr[1])).t())
        cn_b = self.ppr_encoder_cn(torch.stack((cn_ppr[1], cn_ppr[0])).t())
        cn_pe = cn_a + cn_b

        if onehop_ppr is None:
            return cn_pe

        onehop_a = self.ppr_encoder_onehop(
            torch.stack((onehop_ppr[0], onehop_ppr[1])).t())
        onehop_b = self.ppr_encoder_onehop(
            torch.stack((onehop_ppr[1], onehop_ppr[0])).t())
        onehop_pe = onehop_a + onehop_b

        if non1hop_ppr is None:
            return torch.cat((cn_pe, onehop_pe), dim=0)

        non1hop_a = self.ppr_encoder_non1hop(
            torch.stack((non1hop_ppr[0], non1hop_ppr[1])).t())
        non1hop_b = self.ppr_encoder_non1hop(
            torch.stack((non1hop_ppr[1], non1hop_ppr[0])).t())
        non1hop_pe = non1hop_a + non1hop_b

        return torch.cat((cn_pe, onehop_pe, non1hop_pe), dim=0)

    def compute_node_mask(
            self, batch: Tensor, adj: Tensor, ppr_matrix: Tensor
    ) -> Tuple[Tuple, Optional[Tuple], Optional[Tuple]]:
        r"""Get mask based on type of node.

        When mask_type is not "cn", also return the ppr vals for both
        the source and target.

        Args:
            batch (Tensor): The batch vector.
                Denotes which node pairs to predict.
            adj (SparseTensor): Adjacency matrix
            ppr_matrix (Tensor): PPR matrix
        """
        src_adj = torch.index_select(adj, 0, batch[0])
        tgt_adj = torch.index_select(adj, 0, batch[1])

        if self.mask == "cn":
            # 1 when CN, 0 otherwise
            pair_adj = src_adj * tgt_adj
        else:
            # Equals: {0: ">1-Hop", 1: "1-Hop (Non-CN)", 2: "CN"}
            pair_adj = src_adj + tgt_adj

        pair_ix, node_type, src_ppr, tgt_ppr = self.get_ppr_vals(
            batch, pair_adj, ppr_matrix)

        cn_filt_cond = (src_ppr >= self.thresh_cn) & (tgt_ppr
                                                      >= self.thresh_cn)
        onehop_filt_cond = (src_ppr >= self.thresh_1hop) & (
            tgt_ppr >= self.thresh_1hop)

        if self.mask != "cn":
            filt_cond = torch.where(node_type == 1, onehop_filt_cond,
                                    cn_filt_cond)
        else:
            filt_cond = torch.where(node_type == 0, onehop_filt_cond,
                                    cn_filt_cond)

        pair_ix, node_type = pair_ix[:, filt_cond], node_type[filt_cond]
        src_ppr, tgt_ppr = src_ppr[filt_cond], tgt_ppr[filt_cond]

        # >1-Hop mask is gotten separately
        if self.mask == "all":
            non1hop_ix, non1hop_sppr, non1hop_tppr = self.get_non_1hop_ppr(
                batch, adj, ppr_matrix)

        # Dropout
        if self.training and self.trans_drop > 0:
            pair_ix, src_ppr, tgt_ppr, node_type = self.drop_pairwise(
                pair_ix, src_ppr, tgt_ppr, node_type)
            if self.mask == "all":
                non1hop_ix, non1hop_sppr, non1hop_tppr, _ = self.drop_pairwise(
                    non1hop_ix, non1hop_sppr, non1hop_tppr)

        # Separate out CN and 1-Hop
        if self.mask != "cn":
            cn_ind = node_type == 2
            cn_ix = pair_ix[:, cn_ind]
            cn_src_ppr = src_ppr[cn_ind]
            cn_tgt_ppr = tgt_ppr[cn_ind]

            one_hop_ind = node_type == 1
            onehop_ix = pair_ix[:, one_hop_ind]
            onehop_src_ppr = src_ppr[one_hop_ind]
            onehop_tgt_ppr = tgt_ppr[one_hop_ind]

        if self.mask == "cn":
            return (pair_ix, src_ppr, tgt_ppr), None, None
        elif self.mask == "1-hop":
            return (cn_ix, cn_src_ppr, cn_tgt_ppr), (onehop_ix, onehop_src_ppr,
                                                     onehop_tgt_ppr), None
        else:
            return (cn_ix, cn_src_ppr,
                    cn_tgt_ppr), (onehop_ix, onehop_src_ppr,
                                  onehop_tgt_ppr), (non1hop_ix, non1hop_sppr,
                                                    non1hop_tppr)

    def get_ppr_vals(
            self, batch: Tensor, pair_diff_adj: Tensor,
            ppr_matrix: Tensor) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        r"""Get the src and tgt ppr vals.

        Returns the: link the node belongs to, type of node
        (e.g., CN), PPR relative to src, PPR relative to tgt.

        Args:
            batch (Tensor): The batch vector.
                Denotes which node pairs to predict.
            pair_diff_adj (SparseTensor): Combination of rows in
                adjacency for src and tgt nodes (e.g., X1 + X2)
            ppr_matrix (Tensor): PPR matrix
        """
        # Additional terms for also choosing scores when ppr=0
        # Multiplication removes any values for nodes not in batch
        # Addition then adds offset to ensure we select when ppr=0
        # All selected scores are +1 higher than their true val
        src_ppr_adj = torch.index_select(
            ppr_matrix, 0, batch[0]) * pair_diff_adj + pair_diff_adj
        tgt_ppr_adj = torch.index_select(
            ppr_matrix, 0, batch[1]) * pair_diff_adj + pair_diff_adj

        # Can now convert ppr scores to dense
        ppr_ix = src_ppr_adj.coalesce().indices()
        src_ppr = src_ppr_adj.coalesce().values()
        tgt_ppr = tgt_ppr_adj.coalesce().values()

        # TODO: Needed due to a bug in recent torch versions
        # see here for more - https://github.com/pytorch/pytorch/issues/114529
        # note that if one is 0 so is the other
        zero_vals = (src_ppr != 0)
        src_ppr = src_ppr[zero_vals]
        tgt_ppr = tgt_ppr[tgt_ppr != 0]
        ppr_ix = ppr_ix[:, zero_vals]

        pair_diff_adj = pair_diff_adj.coalesce().values()
        node_type = pair_diff_adj[src_ppr != 0]

        # Remove additional +1 from each ppr val
        src_ppr = (src_ppr - node_type) / node_type
        tgt_ppr = (tgt_ppr - node_type) / node_type

        return ppr_ix, node_type, src_ppr, tgt_ppr

    def drop_pairwise(
        self,
        pair_ix: Tensor,
        src_ppr: Optional[Tensor] = None,
        tgt_ppr: Optional[Tensor] = None,
        node_indicator: Optional[Tensor] = None,
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        r"""Perform dropout on pairwise information
        by randomly dropping a percentage of nodes.

        Done before performing attention for efficiency

        Args:
            pair_ix (Tensor): Link node belongs to
            src_ppr (Tensor, optional): PPR relative to src
                (default: :obj:`None`)
            tgt_ppr (Tensor, optional): PPR relative to tgt
                (default: :obj:`None`)
            node_indicator (Tensor, optional): Type of node (e.g., CN)
                (default: :obj:`None`)
        """
        num_indices = math.ceil(pair_ix.size(1) * (1 - self.trans_drop))
        indices = torch.randperm(pair_ix.size(1))[:num_indices]
        pair_ix = pair_ix[:, indices]

        if src_ppr is not None:
            src_ppr = src_ppr[indices]
        if tgt_ppr is not None:
            tgt_ppr = tgt_ppr[indices]
        if node_indicator is not None:
            node_indicator = node_indicator[indices]

        return pair_ix, src_ppr, tgt_ppr, node_indicator

    def get_structure_cnts(
        self,
        batch: Tensor,
        cn_info: Tuple[Tensor, Tensor],
        onehop_info: Tuple[Tensor, Tensor],
        non1hop_info: Optional[Tuple[Tensor, Tensor]],
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        """Counts for CNs, 1-Hop, and >1-Hop that satisfy PPR threshold.

        Also include total # of neighbors

        Args:
            batch (Tensor): The batch vector.
                Denotes which node pairs to predict.
            cn_info (tuple): Information of CN nodes
                Contains (ID of node, src ppr, tgt ppr)
            onehop_info (tuple): Information of 1-Hop nodes.
                Contains (ID of node, src ppr, tgt ppr)
            non1hop_info (tuple): Information of >1-Hop nodes.
                Contains (ID of node, src ppr, tgt ppr)
        """
        num_cns = self.get_num_ppr_thresh(batch, cn_info[0], cn_info[1],
                                          cn_info[2], self.thresh_cn)
        num_1hop = self.get_num_ppr_thresh(batch, onehop_info[0],
                                           onehop_info[1], onehop_info[2],
                                           self.thresh_1hop)

        # TOTAL num of 1-hop neighbors union
        num_ppr_ones = self.get_num_ppr_thresh(batch, onehop_info[0],
                                               onehop_info[1], onehop_info[2],
                                               thresh=0)
        num_neighbors = num_cns + num_ppr_ones

        # Process for >1-hop is different which is why we use get_count below
        if non1hop_info is None:
            return num_cns, num_1hop, 0, num_neighbors
        else:
            num_non1hop = self.get_count(non1hop_info[0], batch)
            return num_cns, num_1hop, num_non1hop, num_neighbors

    def get_num_ppr_thresh(self, batch: Tensor, node_mask: Tensor,
                           src_ppr: Tensor, tgt_ppr: Tensor,
                           thresh: float) -> Tensor:
        """Get # of nodes `v` where `ppr(a, v) >= eta` & `ppr(b, v) >= eta`.

        Args:
            batch (Tensor): The batch vector.
                Denotes which node pairs to predict.
            node_mask (Tensor): IDs of nodes
            src_ppr (Tensor): PPR relative to src node
            tgt_ppr (Tensor): PPR relative to tgt node
            thresh (float): PPR threshold for nodes (`eta`)
        """
        weight = torch.ones(node_mask.size(1), device=node_mask.device)

        ppr_above_thresh = (src_ppr >= thresh) & (tgt_ppr >= thresh)
        num_ppr = scatter(ppr_above_thresh.float() * weight,
                          node_mask[0].long(), dim=0, dim_size=batch.size(1),
                          reduce="sum")
        num_ppr = num_ppr.unsqueeze(-1)

        return num_ppr

    def get_count(
        self,
        node_mask: Tensor,
        batch: Tensor,
    ) -> Tensor:
        """# of nodes for each sample in batch.

        They node have already filtered by PPR beforehand

        Args:
            node_mask (Tensor): IDs of nodes
            batch (Tensor): The batch vector.
                Denotes which node pairs to predict.
        """
        weight = torch.ones(node_mask.size(1), device=node_mask.device)
        num_nodes = scatter(weight, node_mask[0].long(), dim=0,
                            dim_size=batch.size(1), reduce="sum")
        num_nodes = num_nodes.unsqueeze(-1)

        return num_nodes

    def get_non_1hop_ppr(self, batch: Tensor, adj: Tensor,
                         ppr_matrix: Tensor) -> Tensor:
        r"""Get PPR scores for non-1hop nodes.

        Args:
            batch (Tensor): Links in batch
            adj (Tensor): Adjacency matrix
            ppr_matrix (Tensor): Sparse PPR matrix
        """
        # NOTE: Use original adj (one pass in forward() removes links in batch)
        # Done since removing them converts src/tgt nodes to >1-hop nodes.
        # Therefore removing CN and 1-hop will also remove the batch links.

        # During training we add back in the links in the batch
        # (we're removed from adjacency before being passed to model)
        # Done since otherwise they will be mistakenly seen as >1-Hop nodes
        # Instead they're 1-Hop, and get ignored accordingly
        # Ignored during eval since we know the links aren't in the adj
        adj2 = adj
        if self.training:
            n = adj.size(0)
            batch_flip = torch.cat(
                (batch, torch.flip(batch, (0, )).to(batch.device)), dim=-1)
            batch_ones = torch.ones_like(batch_flip[0], device=batch.device)
            adj_edges = torch.sparse_coo_tensor(batch_flip, batch_ones, [n, n],
                                                device=batch.device)
            adj_edges = adj_edges
            adj2 = (adj + adj_edges).coalesce().bool().int()

        src_adj = torch.index_select(adj2, 0, batch[0])
        tgt_adj = torch.index_select(adj2, 0, batch[1])

        src_ppr = torch.index_select(ppr_matrix, 0, batch[0])
        tgt_ppr = torch.index_select(ppr_matrix, 0, batch[1])

        # Remove CN scores
        src_ppr = src_ppr - src_ppr * (src_adj * tgt_adj)
        tgt_ppr = tgt_ppr - tgt_ppr * (src_adj * tgt_adj)
        # Also need to remove CN entries in Adj
        # Otherwise they leak into next computation
        src_adj = src_adj - src_adj * (src_adj * tgt_adj)
        tgt_adj = tgt_adj - tgt_adj * (src_adj * tgt_adj)

        # Remove 1-Hop scores
        src_ppr = src_ppr - src_ppr * (src_adj + tgt_adj)
        tgt_ppr = tgt_ppr - tgt_ppr * (src_adj + tgt_adj)

        # Make sure we include both when we convert to dense so indices align
        # Do so by adding 1 to each based on the other
        src_ppr_add = src_ppr + torch.sign(tgt_ppr)
        tgt_ppr_add = tgt_ppr + torch.sign(src_ppr)

        src_ix = src_ppr_add.coalesce().indices()
        src_vals = src_ppr_add.coalesce().values()
        tgt_vals = tgt_ppr_add.coalesce().values()

        # Now we can remove value which is just 1
        # Technically creates -1 scores for ppr scores that were 0
        # Doesn't matter as they'll be filtered out by condition later
        src_vals = src_vals - 1
        tgt_vals = tgt_vals - 1

        ppr_condition = (src_vals >= self.thresh_non1hop) & (
            tgt_vals >= self.thresh_non1hop)
        src_ix, src_vals, tgt_vals = src_ix[:, ppr_condition], src_vals[
            ppr_condition], tgt_vals[ppr_condition]

        return src_ix, src_vals, tgt_vals

    def calc_sparse_ppr(self, edge_index: Tensor, num_nodes: int,
                        alpha: float = 0.15, eps: float = 5e-5) -> Tensor:
        r"""Calculate the PPR of the graph in sparse format.

        Args:
            edge_index: The edge indices
            num_nodes: Number of nodes
            alpha (float, optional): The alpha value of the PageRank algorithm.
                (default: :obj:`0.15`)
            eps (float, optional): Threshold for stopping the PPR calculation
                (default: :obj:`5e-5`)
        """
        ei, ei_w = get_ppr(edge_index.cpu(), alpha=alpha, eps=eps,
                           num_nodes=num_nodes)
        ppr_matrix = torch.sparse_coo_tensor(ei, ei_w, [num_nodes, num_nodes])

        return ppr_matrix


class LPAttLayer(MessagePassing):
    r"""Attention Layer for pairwise interaction module.

    Args:
        in_channels (int): Size of input dimension
        out_channels (int): Size of output dimension
        node_dim (int): Dimension of nodes being aggregated
        num_heads (int): Number of heads to use in MHA
        dropout (float): Dropout on attention values
        concat (bool, optional): Whether to concat attention
            heads. Otherwise sum (default: :obj:`True`)
    """
    _alpha: OptTensor

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        node_dim: int,
        num_heads: int,
        dropout: float,
        concat: bool = True,
        **kwargs,
    ):
        super().__init__(node_dim=0, flow="target_to_source", **kwargs)

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.heads = num_heads
        self.concat = concat
        self.dropout = dropout
        self.negative_slope = 0.2  # LeakyRelu

        out_dim = 2
        if node_dim is None:
            node_dim = in_channels * out_dim
        else:
            node_dim = node_dim * out_dim

        self.lin_l = Linear(in_channels, self.heads * out_channels,
                            weight_initializer='glorot')
        self.lin_r = Linear(node_dim, self.heads * out_channels,
                            weight_initializer='glorot')

        att_out = out_channels
        self.att = Parameter(Tensor(1, self.heads, att_out))

        if concat:
            self.bias = Parameter(Tensor(self.heads * out_channels))
        else:
            self.bias = Parameter(Tensor(out_channels))

        self._alpha = None

        self.dropout = dropout
        self.post_att_norm = nn.LayerNorm(out_channels)

        self.reset_parameters()

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}({self.in_channels}, '
                f'{self.out_channels}, heads={self.heads})')

    def reset_parameters(self):
        self.lin_l.reset_parameters()
        self.lin_r.reset_parameters()
        self.post_att_norm.reset_parameters()
        glorot(self.att)
        zeros(self.bias)

    def forward(
        self,
        edge_index: Tensor,
        edge_feats: Tensor,
        node_feats: Tensor,
        ppr_rpes: Tensor,
    ) -> Tensor:
        """Runs the forward pass of the module.

        Args:
            edge_index (Tensor): The edge indices.
            edge_feats (Tensor): Concatenated representations
                of src and target nodes for each link
            node_feats (Tensor): Representations for individual
                nodes
            ppr_rpes (Tensor): Relative PEs for each node
        """
        out = self.propagate(edge_index, x=(edge_feats, node_feats),
                             ppr_rpes=ppr_rpes, size=None)

        alpha = self._alpha
        assert alpha is not None
        self._alpha = None

        if self.concat:
            out = out.view(-1, self.heads * self.out_channels)
        else:
            out = out.mean(dim=1)

        if self.bias is not None:
            out = out + self.bias

        out = self.post_att_norm(out)
        out = F.dropout(out, p=self.dropout, training=self.training)

        return out

    def message(self, x_i: Tensor, x_j: Tensor, ppr_rpes: Tensor,
                index: Tensor, ptr: Tensor, size_i: Optional[int]) -> Tensor:
        H, C = self.heads, self.out_channels

        x_j = torch.cat((x_j, ppr_rpes), dim=-1)
        x_j = self.lin_r(x_j).view(-1, H, C)

        # e=(a, b) attending to v
        e1, e2 = x_i.chunk(2, dim=-1)
        e1 = self.lin_l(e1).view(-1, H, C)
        e2 = self.lin_l(e2).view(-1, H, C)
        x = x_j * (e1 + e2)

        x = F.leaky_relu(x, self.negative_slope)
        alpha = (x * self.att).sum(dim=-1)

        alpha = softmax(alpha, index, ptr, size_i)
        self._alpha = alpha

        return x_j * alpha.unsqueeze(-1)


class MLP(nn.Module):
    r"""L Layer MLP."""
    def __init__(self, in_channels: int, hid_channels: int, out_channels: int,
                 num_layers: int = 2, drop: int = 0, norm: str = "layer"):
        super().__init__()
        self.dropout = drop

        if norm == "batch":
            self.norm = nn.BatchNorm1d(hid_channels)
        elif norm == "layer":
            self.norm = nn.LayerNorm(hid_channels)
        else:
            self.norm = None

        self.linears = torch.nn.ModuleList()

        if num_layers == 1:
            self.linears.append(nn.Linear(in_channels, out_channels))
        else:
            self.linears.append(nn.Linear(in_channels, hid_channels))
            for _ in range(num_layers - 2):
                self.linears.append(nn.Linear(hid_channels, hid_channels))
            self.linears.append(nn.Linear(hid_channels, out_channels))

    def reset_parameters(self):
        for lin in self.linears:
            lin.reset_parameters()
        if self.norm is not None:
            self.norm.reset_parameters()

    def forward(self, x: Tensor) -> Tensor:
        for lin in self.linears[:-1]:
            x = lin(x)
            x = self.norm(x) if self.norm is not None else x
            x = F.relu(x)
            x = F.dropout(x, p=self.dropout, training=self.training)

        x = self.linears[-1](x)

        return x.squeeze(-1)



from math import sqrt
from typing import Dict, Optional, Tuple, Union, overload

import torch
from torch import Tensor
from torch.nn.parameter import Parameter

from torch_geometric.explain import (
    ExplainerConfig,
    Explanation,
    HeteroExplanation,
    ModelConfig,
)
from torch_geometric.explain.algorithm import ExplainerAlgorithm
from torch_geometric.explain.algorithm.utils import (
    clear_masks,
    set_hetero_masks,
    set_masks,
)
from torch_geometric.explain.config import MaskType, ModelMode, ModelTaskLevel
from torch_geometric.typing import EdgeType, NodeType


class GNNExplainer(ExplainerAlgorithm):
    r"""The GNN-Explainer model from the `"GNNExplainer: Generating
    Explanations for Graph Neural Networks"
    <https://arxiv.org/abs/1903.03894>`_ paper for identifying compact subgraph
    structures and node features that play a crucial role in the predictions
    made by a GNN.

    .. note::

        For an example of using :class:`GNNExplainer`, see
        `examples/explain/gnn_explainer.py <https://github.com/pyg-team/
        pytorch_geometric/blob/master/examples/explain/gnn_explainer.py>`_,
        `examples/explain/gnn_explainer_ba_shapes.py <https://github.com/
        pyg-team/pytorch_geometric/blob/master/examples/
        explain/gnn_explainer_ba_shapes.py>`_, and `examples/explain/
        gnn_explainer_link_pred.py <https://github.com/pyg-team/
        pytorch_geometric/blob/master/examples/explain/gnn_explainer_link_pred.py>`_.

    .. note::

        The :obj:`edge_size` coefficient is multiplied by the number of nodes
        in the explanation at every iteration, and the resulting value is added
        to the loss as a regularization term, with the goal of producing
        compact explanations.
        A higher value will push the algorithm towards explanations with less
        elements.
        Consider adjusting the :obj:`edge_size` coefficient according to the
        average node degree in the dataset, especially if this value is bigger
        than in the datasets used in the original paper.

    Args:
        epochs (int, optional): The number of epochs to train.
            (default: :obj:`100`)
        lr (float, optional): The learning rate to apply.
            (default: :obj:`0.01`)
        **kwargs (optional): Additional hyper-parameters to override default
            settings in
            :attr:`~torch_geometric.explain.algorithm.GNNExplainer.coeffs`.
    """

    default_coeffs = {
        'edge_size': 0.005,
        'edge_reduction': 'sum',
        'node_feat_size': 1.0,
        'node_feat_reduction': 'mean',
        'edge_ent': 1.0,
        'node_feat_ent': 0.1,
        'EPS': 1e-15,
    }

    def __init__(self, epochs: int = 100, lr: float = 0.01, **kwargs):
        super().__init__()
        self.epochs = epochs
        self.lr = lr
        self.coeffs = dict(self.default_coeffs)
        self.coeffs.update(kwargs)

        self.node_mask = self.hard_node_mask = None
        self.edge_mask = self.hard_edge_mask = None
        self.is_hetero = False

    @overload
    def forward(
        self,
        model: torch.nn.Module,
        x: Tensor,
        edge_index: Tensor,
        *,
        target: Tensor,
        index: Optional[Union[int, Tensor]] = None,
        **kwargs,
    ) -> Explanation:
        ...

    @overload
    def forward(
        self,
        model: torch.nn.Module,
        x: Dict[NodeType, Tensor],
        edge_index: Dict[EdgeType, Tensor],
        *,
        target: Tensor,
        index: Optional[Union[int, Tensor]] = None,
        **kwargs,
    ) -> HeteroExplanation:
        ...

    def forward(
        self,
        model: torch.nn.Module,
        x: Union[Tensor, Dict[NodeType, Tensor]],
        edge_index: Union[Tensor, Dict[EdgeType, Tensor]],
        *,
        target: Tensor,
        index: Optional[Union[int, Tensor]] = None,
        **kwargs,
    ) -> Union[Explanation, HeteroExplanation]:
        self.is_hetero = isinstance(x, dict)
        self._train(model, x, edge_index, target=target, index=index, **kwargs)
        explanation = self._create_explanation()
        self._clean_model(model)
        return explanation

    def _create_explanation(self) -> Union[Explanation, HeteroExplanation]:
        """Create an explanation object from the current masks."""
        if self.is_hetero:
            # For heterogeneous graphs, process each type separately
            node_mask_dict = {}
            edge_mask_dict = {}

            for node_type, mask in self.node_mask.items():
                if mask is not None:
                    node_mask_dict[node_type] = self._post_process_mask(
                        mask,
                        self.hard_node_mask[node_type],
                        apply_sigmoid=True,
                    )

            for edge_type, mask in self.edge_mask.items():
                if mask is not None:
                    edge_mask_dict[edge_type] = self._post_process_mask(
                        mask,
                        self.hard_edge_mask[edge_type],
                        apply_sigmoid=True,
                    )

            # Create heterogeneous explanation
            explanation = HeteroExplanation()
            explanation.set_value_dict('node_mask', node_mask_dict)
            explanation.set_value_dict('edge_mask', edge_mask_dict)

        else:
            # For homogeneous graphs, process single masks
            node_mask = self._post_process_mask(
                self.node_mask,
                self.hard_node_mask,
                apply_sigmoid=True,
            )
            edge_mask = self._post_process_mask(
                self.edge_mask,
                self.hard_edge_mask,
                apply_sigmoid=True,
            )

            # Create homogeneous explanation
            explanation = Explanation(node_mask=node_mask, edge_mask=edge_mask)

        return explanation

    def supports(self) -> bool:
        return True

    @overload
    def _train(
        self,
        model: torch.nn.Module,
        x: Tensor,
        edge_index: Tensor,
        *,
        target: Tensor,
        index: Optional[Union[int, Tensor]] = None,
        **kwargs,
    ) -> None:
        ...

    @overload
    def _train(
        self,
        model: torch.nn.Module,
        x: Dict[NodeType, Tensor],
        edge_index: Dict[EdgeType, Tensor],
        *,
        target: Tensor,
        index: Optional[Union[int, Tensor]] = None,
        **kwargs,
    ) -> None:
        ...

    def _train(
        self,
        model: torch.nn.Module,
        x: Union[Tensor, Dict[NodeType, Tensor]],
        edge_index: Union[Tensor, Dict[EdgeType, Tensor]],
        *,
        target: Tensor,
        index: Optional[Union[int, Tensor]] = None,
        **kwargs,
    ) -> None:
        # Initialize masks based on input type
        self._initialize_masks(x, edge_index)

        # Collect parameters for optimization
        parameters = self._collect_parameters(model, edge_index)

        # Create optimizer
        optimizer = torch.optim.Adam(parameters, lr=self.lr)

        # Training loop
        for i in range(self.epochs):
            optimizer.zero_grad()

            # Forward pass with masked inputs
            y_hat = self._forward_with_masks(model, x, edge_index, **kwargs)
            y = target

            # Handle index if provided
            if index is not None:
                y_hat, y = y_hat[index], y[index]

            # Calculate loss
            loss = self._loss(y_hat, y)

            # Backward pass
            loss.backward()
            optimizer.step()

            # In the first iteration, collect gradients to identify important
            # nodes/edges
            if i == 0:
                self._collect_gradients()

    def _collect_parameters(self, model, edge_index):
        """Collect parameters for optimization."""
        parameters = []

        if self.is_hetero:
            # For heterogeneous graphs, collect parameters from all types
            for mask in self.node_mask.values():
                if mask is not None:
                    parameters.append(mask)
            if any(v is not None for v in self.edge_mask.values()):
                set_hetero_masks(model, self.edge_mask, edge_index)
            for mask in self.edge_mask.values():
                if mask is not None:
                    parameters.append(mask)
        else:
            # For homogeneous graphs, collect single parameters
            if self.node_mask is not None:
                parameters.append(self.node_mask)
            if self.edge_mask is not None:
                set_masks(model, self.edge_mask, edge_index,
                          apply_sigmoid=True)
                parameters.append(self.edge_mask)

        return parameters

    @overload
    def _forward_with_masks(
        self,
        model: torch.nn.Module,
        x: Tensor,
        edge_index: Tensor,
        **kwargs,
    ) -> Tensor:
        ...

    @overload
    def _forward_with_masks(
        self,
        model: torch.nn.Module,
        x: Dict[NodeType, Tensor],
        edge_index: Dict[EdgeType, Tensor],
        **kwargs,
    ) -> Tensor:
        ...

    def _forward_with_masks(
        self,
        model: torch.nn.Module,
        x: Union[Tensor, Dict[NodeType, Tensor]],
        edge_index: Union[Tensor, Dict[EdgeType, Tensor]],
        **kwargs,
    ) -> Tensor:
        """Forward pass with masked inputs."""
        if self.is_hetero:
            # Apply masks to heterogeneous inputs
            h_dict = {}
            for node_type, features in x.items():
                if node_type in self.node_mask and self.node_mask[
                        node_type] is not None:
                    h_dict[node_type] = features * self.node_mask[
                        node_type].sigmoid()
                else:
                    h_dict[node_type] = features

            # Forward pass with masked features
            return model(h_dict, edge_index, **kwargs)
        else:
            # Apply mask to homogeneous input
            h = x if self.node_mask is None else x * self.node_mask.sigmoid()

            # Forward pass with masked features
            return model(h, edge_index, **kwargs)

    def _initialize_masks(
        self,
        x: Union[Tensor, Dict[NodeType, Tensor]],
        edge_index: Union[Tensor, Dict[EdgeType, Tensor]],
    ) -> None:
        node_mask_type = self.explainer_config.node_mask_type
        edge_mask_type = self.explainer_config.edge_mask_type

        if self.is_hetero:
            # Initialize dictionaries for heterogeneous masks
            self.node_mask = {}
            self.hard_node_mask = {}
            self.edge_mask = {}
            self.hard_edge_mask = {}

            # Initialize node masks for each node type
            for node_type, features in x.items():
                device = features.device
                N, F = features.size()
                self._initialize_node_mask(node_mask_type, node_type, N, F,
                                           device)

            # Initialize edge masks for each edge type
            for edge_type, indices in edge_index.items():
                device = indices.device
                E = indices.size(1)
                N = max(indices.max().item() + 1,
                        max(feat.size(0) for feat in x.values()))
                self._initialize_edge_mask(edge_mask_type, edge_type, E, N,
                                           device)
        else:
            # Initialize masks for homogeneous graph
            device = x.device
            (N, F), E = x.size(), edge_index.size(1)

            # Initialize homogeneous node and edge masks
            self._initialize_homogeneous_masks(node_mask_type, edge_mask_type,
                                               N, F, E, device)

    def _initialize_node_mask(
        self,
        node_mask_type,
        node_type,
        N,
        F,
        device,
    ) -> None:
        """Initialize node mask for a specific node type."""
        std = 0.1
        if node_mask_type is None:
            self.node_mask[node_type] = None
            self.hard_node_mask[node_type] = None
        elif node_mask_type == MaskType.object:
            self.node_mask[node_type] = Parameter(
                torch.randn(N, 1, device=device) * std)
            self.hard_node_mask[node_type] = None
        elif node_mask_type == MaskType.attributes:
            self.node_mask[node_type] = Parameter(
                torch.randn(N, F, device=device) * std)
            self.hard_node_mask[node_type] = None
        elif node_mask_type == MaskType.common_attributes:
            self.node_mask[node_type] = Parameter(
                torch.randn(1, F, device=device) * std)
            self.hard_node_mask[node_type] = None
        else:
            raise ValueError(f"Invalid node mask type: {node_mask_type}")

    def _initialize_edge_mask(self, edge_mask_type, edge_type, E, N, device):
        """Initialize edge mask for a specific edge type."""
        if edge_mask_type is None:
            self.edge_mask[edge_type] = None
            self.hard_edge_mask[edge_type] = None
        elif edge_mask_type == MaskType.object:
            std = torch.nn.init.calculate_gain('relu') * sqrt(2.0 / (2 * N))
            self.edge_mask[edge_type] = Parameter(
                torch.randn(E, device=device) * std)
            self.hard_edge_mask[edge_type] = None
        else:
            raise ValueError(f"Invalid edge mask type: {edge_mask_type}")

    def _initialize_homogeneous_masks(self, node_mask_type, edge_mask_type, N,
                                      F, E, device):
        """Initialize masks for homogeneous graph."""
        # Initialize node mask
        std = 0.1
        if node_mask_type is None:
            self.node_mask = None
        elif node_mask_type == MaskType.object:
            self.node_mask = Parameter(torch.randn(N, 1, device=device) * std)
        elif node_mask_type == MaskType.attributes:
            self.node_mask = Parameter(torch.randn(N, F, device=device) * std)
        elif node_mask_type == MaskType.common_attributes:
            self.node_mask = Parameter(torch.randn(1, F, device=device) * std)
        else:
            raise ValueError(f"Invalid node mask type: {node_mask_type}")

        # Initialize edge mask
        if edge_mask_type is None:
            self.edge_mask = None
        elif edge_mask_type == MaskType.object:
            std = torch.nn.init.calculate_gain('relu') * sqrt(2.0 / (2 * N))
            self.edge_mask = Parameter(torch.randn(E, device=device) * std)
        else:
            raise ValueError(f"Invalid edge mask type: {edge_mask_type}")

    def _collect_gradients(self) -> None:
        if self.is_hetero:
            self._collect_hetero_gradients()
        else:
            self._collect_homo_gradients()

    def _collect_hetero_gradients(self):
        """Collect gradients for heterogeneous graph."""
        for node_type, mask in self.node_mask.items():
            if mask is not None:
                if mask.grad is None:
                    raise ValueError(
                        f"Could not compute gradients for node masks of type "
                        f"'{node_type}'. Please make sure that node masks are "
                        f"used inside the model or disable it via "
                        f"`node_mask_type=None`.")

                self.hard_node_mask[node_type] = mask.grad != 0.0

        for edge_type, mask in self.edge_mask.items():
            if mask is not None:
                if mask.grad is None:
                    raise ValueError(
                        f"Could not compute gradients for edge masks of type "
                        f"'{edge_type}'. Please make sure that edge masks are "
                        f"used inside the model or disable it via "
                        f"`edge_mask_type=None`.")
                self.hard_edge_mask[edge_type] = mask.grad != 0.0

    def _collect_homo_gradients(self):
        """Collect gradients for homogeneous graph."""
        if self.node_mask is not None:
            if self.node_mask.grad is None:
                raise ValueError("Could not compute gradients for node "
                                 "features. Please make sure that node "
                                 "features are used inside the model or "
                                 "disable it via `node_mask_type=None`.")
            self.hard_node_mask = self.node_mask.grad != 0.0

        if self.edge_mask is not None:
            if self.edge_mask.grad is None:
                raise ValueError("Could not compute gradients for edges. "
                                 "Please make sure that edges are used "
                                 "via message passing inside the model or "
                                 "disable it via `edge_mask_type=None`.")
            self.hard_edge_mask = self.edge_mask.grad != 0.0

    def _loss(self, y_hat: Tensor, y: Tensor) -> Tensor:
        # Calculate base loss based on model configuration
        loss = self._calculate_base_loss(y_hat, y)

        # Apply regularization based on graph type
        if self.is_hetero:
            # Apply regularization for heterogeneous graph
            loss = self._apply_hetero_regularization(loss)
        else:
            # Apply regularization for homogeneous graph
            loss = self._apply_homo_regularization(loss)

        return loss

    def _calculate_base_loss(self, y_hat, y):
        """Calculate base loss based on model configuration."""
        if self.model_config.mode == ModelMode.binary_classification:
            return self._loss_binary_classification(y_hat, y)
        elif self.model_config.mode == ModelMode.multiclass_classification:
            return self._loss_multiclass_classification(y_hat, y)
        elif self.model_config.mode == ModelMode.regression:
            return self._loss_regression(y_hat, y)
        else:
            raise ValueError(f"Invalid model mode: {self.model_config.mode}")

    def _apply_hetero_regularization(self, loss):
        """Apply regularization for heterogeneous graph."""
        # Apply regularization for each edge type
        for edge_type, mask in self.edge_mask.items():
            if (mask is not None
                    and self.hard_edge_mask[edge_type] is not None):
                loss = self._add_mask_regularization(
                    loss, mask, self.hard_edge_mask[edge_type],
                    self.coeffs['edge_size'], self.coeffs['edge_reduction'],
                    self.coeffs['edge_ent'])

        # Apply regularization for each node type
        for node_type, mask in self.node_mask.items():
            if (mask is not None
                    and self.hard_node_mask[node_type] is not None):
                loss = self._add_mask_regularization(
                    loss, mask, self.hard_node_mask[node_type],
                    self.coeffs['node_feat_size'],
                    self.coeffs['node_feat_reduction'],
                    self.coeffs['node_feat_ent'])

        return loss

    def _apply_homo_regularization(self, loss):
        """Apply regularization for homogeneous graph."""
        # Apply regularization for edge mask
        if self.hard_edge_mask is not None:
            assert self.edge_mask is not None
            loss = self._add_mask_regularization(loss, self.edge_mask,
                                                 self.hard_edge_mask,
                                                 self.coeffs['edge_size'],
                                                 self.coeffs['edge_reduction'],
                                                 self.coeffs['edge_ent'])

        # Apply regularization for node mask
        if self.hard_node_mask is not None:
            assert self.node_mask is not None
            loss = self._add_mask_regularization(
                loss, self.node_mask, self.hard_node_mask,
                self.coeffs['node_feat_size'],
                self.coeffs['node_feat_reduction'],
                self.coeffs['node_feat_ent'])

        return loss

    def _add_mask_regularization(self, loss, mask, hard_mask, size_coeff,
                                 reduction_name, ent_coeff):
        """Add size and entropy regularization for a mask."""
        m = mask[hard_mask].sigmoid()
        reduce_fn = getattr(torch, reduction_name)
        # Add size regularization
        loss = loss + size_coeff * reduce_fn(m)
        # Add entropy regularization
        ent = -m * torch.log(m + self.coeffs['EPS']) - (
            1 - m) * torch.log(1 - m + self.coeffs['EPS'])
        loss = loss + ent_coeff * ent.mean()

        return loss

    def _clean_model(self, model):
        clear_masks(model)
        self.node_mask = self.hard_node_mask = None
        self.edge_mask = self.hard_edge_mask = None


class GNNExplainer_:
    r"""Deprecated version for :class:`GNNExplainer`."""

    coeffs = GNNExplainer.default_coeffs

    conversion_node_mask_type = {
        'feature': 'common_attributes',
        'individual_feature': 'attributes',
        'scalar': 'object',
    }

    conversion_return_type = {
        'log_prob': 'log_probs',
        'prob': 'probs',
        'raw': 'raw',
        'regression': 'raw',
    }

    def __init__(
        self,
        model: torch.nn.Module,
        epochs: int = 100,
        lr: float = 0.01,
        return_type: str = 'log_prob',
        feat_mask_type: str = 'feature',
        allow_edge_mask: bool = True,
        **kwargs,
    ):
        assert feat_mask_type in ['feature', 'individual_feature', 'scalar']

        explainer_config = ExplainerConfig(
            explanation_type='model',
            node_mask_type=self.conversion_node_mask_type[feat_mask_type],
            edge_mask_type=MaskType.object if allow_edge_mask else None,
        )
        model_config = ModelConfig(
            mode='regression'
            if return_type == 'regression' else 'multiclass_classification',
            task_level=ModelTaskLevel.node,
            return_type=self.conversion_return_type[return_type],
        )

        self.model = model
        self._explainer = GNNExplainer(epochs=epochs, lr=lr, **kwargs)
        self._explainer.connect(explainer_config, model_config)

    @torch.no_grad()
    def get_initial_prediction(self, *args, **kwargs) -> Tensor:

        training = self.model.training
        self.model.eval()

        out = self.model(*args, **kwargs)
        if (self._explainer.model_config.mode ==
                ModelMode.multiclass_classification):
            out = out.argmax(dim=-1)

        self.model.train(training)

        return out

    def explain_graph(
        self,
        x: Tensor,
        edge_index: Tensor,
        **kwargs,
    ) -> Tuple[Tensor, Tensor]:
        self._explainer.model_config.task_level = ModelTaskLevel.graph

        explanation = self._explainer(
            self.model,
            x,
            edge_index,
            target=self.get_initial_prediction(x, edge_index, **kwargs),
            **kwargs,
        )
        return self._convert_output(explanation, edge_index)

    def explain_node(
        self,
        node_idx: int,
        x: Tensor,
        edge_index: Tensor,
        **kwargs,
    ) -> Tuple[Tensor, Tensor]:
        self._explainer.model_config.task_level = ModelTaskLevel.node
        explanation = self._explainer(
            self.model,
            x,
            edge_index,
            target=self.get_initial_prediction(x, edge_index, **kwargs),
            index=node_idx,
            **kwargs,
        )
        return self._convert_output(explanation, edge_index, index=node_idx,
                                    x=x)

    def _convert_output(self, explanation, edge_index, index=None, x=None):
        node_mask = explanation.get('node_mask')
        edge_mask = explanation.get('edge_mask')

        if node_mask is not None:
            node_mask_type = self._explainer.explainer_config.node_mask_type
            if node_mask_type in {MaskType.object, MaskType.common_attributes}:
                node_mask = node_mask.view(-1)

        if edge_mask is None:
            if index is not None:
                _, edge_mask = self._explainer._get_hard_masks(
                    self.model, index, edge_index, num_nodes=x.size(0))
                edge_mask = edge_mask.to(x.dtype)
            else:
                edge_mask = torch.ones(edge_index.size(1),
                                       device=edge_index.device)

        return node_mask, edge_mask



import os.path as osp
from typing import Callable, Optional

import numpy as np
import torch

from torch_geometric.data import Data, InMemoryDataset, download_url


class GemsecDeezer(InMemoryDataset):
    r"""The Deezer User Network datasets introduced in the
    `"GEMSEC: Graph Embedding with Self Clustering"
    <https://arxiv.org/abs/1802.03997>`_ paper.
    Nodes represent Deezer user and edges are mutual friendships.
    The task is multi-label multi-class node classification about
    the genres liked by the users.

    Args:
        root (str): Root directory where the dataset should be saved.
        name (str): The name of the dataset (:obj:`"HU"`, :obj:`"HR"`,
            :obj:`"RO"`).
        transform (callable, optional): A function/transform that takes in an
            :obj:`torch_geometric.data.Data` object and returns a transformed
            version. The data object will be transformed before every access.
            (default: :obj:`None`)
        pre_transform (callable, optional): A function/transform that takes in
            an :obj:`torch_geometric.data.Data` object and returns a
            transformed version. The data object will be transformed before
            being saved to disk. (default: :obj:`None`)
        force_reload (bool, optional): Whether to re-process the dataset.
            (default: :obj:`False`)
    """

    url = 'https://graphmining.ai/datasets/ptg/gemsec'

    def __init__(
        self,
        root: str,
        name: str,
        transform: Optional[Callable] = None,
        pre_transform: Optional[Callable] = None,
        force_reload: bool = False,
    ) -> None:
        self.name = name
        assert self.name in ['HU', 'HR', 'RO']
        super().__init__(root, transform, pre_transform,
                         force_reload=force_reload)
        self.load(self.processed_paths[0])

    @property
    def raw_dir(self) -> str:
        return osp.join(self.root, self.name, 'raw')

    @property
    def processed_dir(self) -> str:
        return osp.join(self.root, self.name, 'processed')

    @property
    def raw_file_names(self) -> str:
        return f'{self.name}.npz'

    @property
    def processed_file_names(self) -> str:
        return 'data.pt'

    def download(self) -> None:
        download_url(osp.join(self.url, self.name + '.npz'), self.raw_dir)

    def process(self) -> None:
        data = np.load(self.raw_paths[0], 'r', allow_pickle=True)
        y = torch.from_numpy(data['target']).to(torch.long)
        edge_index = torch.from_numpy(data['edges']).to(torch.long)
        edge_index = edge_index.t().contiguous()

        data = Data(y=y, edge_index=edge_index)

        if self.pre_transform is not None:
            data = self.pre_transform(data)

        self.save([data], self.processed_paths[0])



import torch
from torch import Tensor


class Reshape(torch.nn.Module):
    def __init__(self, *shape):
        super().__init__()
        self.shape = shape

    def forward(self, x: Tensor) -> Tensor:
        """"""  # noqa: D419
        x = x.view(*self.shape)
        return x

    def __repr__(self) -> str:
        shape = ', '.join([str(dim) for dim in self.shape])
        return f'{self.__class__.__name__}({shape})'



from torch_geometric.deprecation import deprecated

from .pgm_explainer import PGMExplainer
from torch_geometric.explain.algorithm.graphmask_explainer import (
    GraphMaskExplainer as NewGraphMaskExplainer)

GraphMaskExplainer = deprecated(
    "use 'torch_geometric.explain.algorithm.GraphMaskExplainer' instead", )(
        NewGraphMaskExplainer)

__all__ = classes = [
    'PGMExplainer',
]



import warnings
from os import PathLike
from typing import Any, Union

import torch

from torch_geometric import is_compiling


def is_in_onnx_export() -> bool:
    r"""Returns :obj:`True` in case :pytorch:`PyTorch` is exporting to ONNX via
    :meth:`torch.onnx.export`.
    """
    if is_compiling():
        return False
    if torch.jit.is_scripting():
        return False
    return torch.onnx.is_in_onnx_export()


def safe_onnx_export(
    model: torch.nn.Module,
    args: Union[torch.Tensor, tuple[Any, ...]],
    f: Union[str, PathLike[Any], None],
    skip_on_error: bool = False,
    **kwargs: Any,
) -> bool:
    r"""A safe wrapper around :meth:`torch.onnx.export` that handles known
    ONNX serialization issues in PyTorch Geometric.

    This function provides workarounds for the ``onnx_ir.serde.SerdeError``
    with boolean ``allowzero`` attributes that occurs in certain environments.

    Args:
        model (torch.nn.Module): The model to export.
        args (torch.Tensor or tuple): The input arguments for the model.
        f (str or PathLike): The file path to save the model.
        skip_on_error (bool): If True, return False instead of raising when
            workarounds fail. Useful for CI environments.
        **kwargs: Additional arguments passed to :meth:`torch.onnx.export`.

    Returns:
        bool: True if export succeeded, False if skipped due to known issues
              (only when skip_on_error=True).

    Example:
        >>> from torch_geometric.nn import SAGEConv
        >>> from torch_geometric import safe_onnx_export
        >>>
        >>> class MyModel(torch.nn.Module):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.conv = SAGEConv(8, 16)
        ...     def forward(self, x, edge_index):
        ...         return self.conv(x, edge_index)
        >>>
        >>> model = MyModel()
        >>> x = torch.randn(3, 8)
        >>> edge_index = torch.tensor([[0, 1, 2], [1, 0, 2]])
        >>> success = safe_onnx_export(model, (x, edge_index), 'model.onnx')
        >>>
        >>> # For CI environments:
        >>> success = safe_onnx_export(model, (x, edge_index), 'model.onnx',
        ...                             skip_on_error=True)
        >>> if not success:
        ...     print("ONNX export skipped due to known upstream issue")
    """
    # Convert single tensor to tuple for torch.onnx.export compatibility
    if isinstance(args, torch.Tensor):
        args = (args, )

    try:
        # First attempt: standard ONNX export
        torch.onnx.export(model, args, f, **kwargs)
        return True

    except Exception as e:
        error_str = str(e)
        error_type = type(e).__name__

        # Check for the specific onnx_ir.serde.SerdeError patterns
        is_allowzero_error = (('onnx_ir.serde.SerdeError' in error_str
                               and 'allowzero' in error_str) or
                              'ValueError: Value out of range: 1' in error_str
                              or 'serialize_model_into' in error_str
                              or 'serialize_attribute_into' in error_str)

        if is_allowzero_error:
            warnings.warn(
                f"Encountered known ONNX serialization issue ({error_type}). "
                "This is likely the allowzero boolean attribute bug. "
                "Attempting workaround...", UserWarning, stacklevel=2)

            # Apply workaround strategies
            return _apply_onnx_allowzero_workaround(model, args, f,
                                                    skip_on_error, **kwargs)

        else:
            # Re-raise other errors
            raise


def _apply_onnx_allowzero_workaround(
    model: torch.nn.Module,
    args: tuple[Any, ...],
    f: Union[str, PathLike[Any], None],
    skip_on_error: bool = False,
    **kwargs: Any,
) -> bool:
    r"""Apply workaround strategies for onnx_ir.serde.SerdeError with allowzero
    attributes.

    Returns:
        bool: True if export succeeded, False if skipped (when
              skip_on_error=True).
    """
    # Strategy 1: Try without dynamo if it was enabled
    if kwargs.get('dynamo', False):
        try:
            kwargs_no_dynamo = kwargs.copy()
            kwargs_no_dynamo['dynamo'] = False

            warnings.warn(
                "Retrying ONNX export with dynamo=False as workaround",
                UserWarning, stacklevel=3)

            torch.onnx.export(model, args, f, **kwargs_no_dynamo)
            return True

        except Exception:
            pass

    # Strategy 2: Try with different opset versions
    original_opset = kwargs.get('opset_version', 18)
    for opset_version in [17, 16, 15, 14, 13, 11]:
        if opset_version != original_opset:
            try:
                kwargs_opset = kwargs.copy()
                kwargs_opset['opset_version'] = opset_version

                warnings.warn(
                    f"Retrying ONNX export with opset_version={opset_version}",
                    UserWarning, stacklevel=3)

                torch.onnx.export(model, args, f, **kwargs_opset)
                return True

            except Exception:
                continue

    # Strategy 3: Try legacy export (non-dynamo with older opset)
    try:
        kwargs_legacy = kwargs.copy()
        kwargs_legacy['dynamo'] = False
        kwargs_legacy['opset_version'] = 11

        warnings.warn(
            "Retrying ONNX export with legacy settings "
            "(dynamo=False, opset_version=11)", UserWarning, stacklevel=3)

        torch.onnx.export(model, args, f, **kwargs_legacy)
        return True

    except Exception:
        pass

    # Strategy 4: Try with minimal settings
    try:
        minimal_kwargs: dict[str, Any] = {
            'opset_version': 11,
            'dynamo': False,
        }
        # Add optional parameters if they exist
        if kwargs.get('input_names') is not None:
            minimal_kwargs['input_names'] = kwargs.get('input_names')
        if kwargs.get('output_names') is not None:
            minimal_kwargs['output_names'] = kwargs.get('output_names')

        warnings.warn(
            "Retrying ONNX export with minimal settings as last resort",
            UserWarning, stacklevel=3)

        torch.onnx.export(model, args, f, **minimal_kwargs)
        return True

    except Exception:
        pass

    # If all strategies fail, handle based on skip_on_error flag
    import os
    pytest_detected = 'PYTEST_CURRENT_TEST' in os.environ or 'pytest' in str(f)

    if skip_on_error:
        # For CI environments: skip gracefully instead of failing
        warnings.warn(
            "ONNX export skipped due to known upstream issue "
            "(onnx_ir.serde.SerdeError). "
            "This is caused by a bug in the onnx_ir package where boolean "
            "allowzero attributes cannot be serialized. All workarounds "
            "failed. Consider updating packages: pip install --upgrade onnx "
            "onnxscript "
            "onnx_ir", UserWarning, stacklevel=3)
        return False

    # For regular usage: provide detailed error message
    error_msg = (
        "Failed to export model to ONNX due to known serialization issue. "
        "This is caused by a bug in the onnx_ir package where boolean "
        "allowzero attributes cannot be serialized. "
        "Workarounds attempted: dynamo=False, multiple opset versions, "
        "and legacy export. ")

    if pytest_detected:
        error_msg += (
            "\n\nThis error commonly occurs in pytest environments. "
            "Try one of these solutions:\n"
            "1. Run the export outside of pytest (in a regular Python "
            "script)\n"
            "2. Update packages: pip install --upgrade onnx onnxscript "
            "onnx_ir\n"
            "3. Use torch.jit.script() instead of ONNX export for testing\n"
            "4. Use safe_onnx_export(..., skip_on_error=True) to skip "
            "gracefully in CI")
    else:
        error_msg += ("\n\nTry updating packages: pip install --upgrade onnx "
                      "onnxscript onnx_ir")

    raise RuntimeError(error_msg)



import warnings

import torch

try:
    import lightning.pytorch as pl
    _pl_is_available = True
except ImportError:
    try:
        import pytorch_lightning as pl
        _pl_is_available = True
    except ImportError:
        _pl_is_available = False

if _pl_is_available:
    LightningModule = pl.LightningModule
    Callback = pl.Callback
else:
    pl = object
    LightningModule = torch.nn.Module
    Callback = object

    warnings.warn(
        "To use GraphGym, install 'pytorch_lightning' or 'lightning' via "
        "'pip install pytorch_lightning' or 'pip install lightning'",
        stacklevel=2)
